import logging
import os
from pathlib import Path
from uuid import UUID

import orjson
from beartype import beartype

import picsellia.pxl_multithreading as mlt
from picsellia import exceptions
from picsellia.colors import Colors
from picsellia.compatibility import add_data_mandatory_query_parameters
from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    BadRequestError,
    NoConnectorFound,
    NoDataError,
    NothingDoneError,
    UnprocessableData,
)
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data, MultiData
from picsellia.sdk.data_projection import DataProjection
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.job import Job
from picsellia.sdk.processing import Processing
from picsellia.sdk.tag import Tag
from picsellia.services.data_uploader import DataUploader
from picsellia.services.datasource import DataSourceService
from picsellia.services.error_manager import ErrorManager
from picsellia.services.lister.data_lister import DataFilter, DataLister
from picsellia.services.visual_search import DatalakeVisualSearchService
from picsellia.types.enums import DataUploadStatus, ObjectDataType, TagTarget
from picsellia.types.schemas import CloudObject, CloudProjectionObject, DatalakeSchema
from picsellia.utils import (
    combine_two_ql,
    convert_tag_list_to_query_language,
    filter_payload,
)

logger = logging.getLogger("picsellia")


class Datalake(Dao):
    def __init__(self, connexion: Connection, organization_id: UUID, data: dict):
        Dao.__init__(self, connexion, data)
        self._organization_id = organization_id

    def __str__(self):
        return f"{Colors.GREEN}Datalake '{self.name}'{Colors.ENDC} (id: {self.id})"

    @property
    def name(self) -> str:
        """Name of this (Datalake)"""
        return self._name

    @property
    def connector_id(self) -> UUID:
        """Connector id used by this (Datalake)"""
        if self._connector_id is None:
            raise NoConnectorFound(
                "This datalake has no connector. You cannot retrieve and upload data into this datalake."
            )
        return self._connector_id

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = DatalakeSchema(**data)
        self._name = schema.name
        if schema.connector_id is not None:
            self._connector_id = schema.connector_id
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/datalake/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/datalake/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/datalake/{self.id}"

    @exception_handler
    @beartype
    def upload_data(  # noqa: C901
        self,
        filepaths: str | Path | list[str | Path],
        tags: list[str | Tag] | None = None,
        source: str | DataSource | None = None,
        max_workers: int | None = None,
        error_manager: ErrorManager | None = None,
        metadata: None | dict | list[dict] = None,
        fill_metadata: bool | None = False,
        wait_for_unprocessable_data: bool | None = True,
        upload_dir: str | None = None,
        custom_metadata: None | dict | list[dict] = None,
    ) -> Data | MultiData:
        """Upload data into this datalake.

        Upload files representing data, into a datalake.
        You can give some tags as a list.
        You can give a source for your data.

        If some data fails to upload, check the example to see how
        to retrieve the list of file paths that failed.

        For more information about metadata, check https://documentation.picsellia.com/docs/metadata

        Examples:
            ```python
            from picsellia.services.error_manager import ErrorManager

            source_camera_one = client.get_datasource("camera-one")
            source_camera_two = client.get_datasource("camera-two")

            lake = client.get_datalake()

            tag_car = lake.get_data_tag("car")
            tag_huge_car = lake.get_data_tag("huge-car")

            lake.upload_data(filepaths=["porsche.png", "ferrari.png"], tags=[tag_car], source=source_camera_one)
            lake.upload_data(filepaths="truck.png", tags=[tag_huge_car], source=source_camera_two, metadata={"longitude": 43.6027273, "latitude": 1.4541129}, fill_metadata=True)

            error_manager = ErrorManager()
            lake.upload_data(filepaths=["twingo.png", "path/unknown.png", error_manager=error_manager)

            # This call will return a list of UploadError to see what was wrong
            error_paths = [error.path for error in error_manager.errors]
            ```
        Arguments:
            filepaths (str or Path or list[str or Path]): Filepaths of your data
            tags (list[Tag], optional): Data Tags that will be given to data. Defaults to [].
            source (DataSource, optional): Source of your data.
            max_workers (int, optional): Number of max workers used to upload. Defaults to os.cpu_count() + 4.
            error_manager (ErrorManager, optional): Giving an ErrorManager will allow you to retrieve errors
            metadata (Dict or list[Dict], optional): Add some metadata to given data, filepaths length must match
                 this parameter. Defaults to no metadata.
            fill_metadata (bool, optional): Whether read exif tags of image and add it into metadata field.
                 If some fields are already given in metadata fields, they will be overridden.
            wait_for_unprocessable_data (bool, optional): If true, this method will wait for all data to be fully
                uploaded and processed by our services. Defaults to true.
            upload_dir (str, optional): This parameter can only be used with private object-storages. Specify this parameter to prefix the object name of the data. Filename will still contain a generated uuid4
            custom_metadata (Dict or list[Dict], optional): Add custom metadata to given data, filepaths length must match this parameter. Defaults to no custom metadata.
        Returns:
            A (Data) object or a (MultiData) object that wraps a list of Data.
        """
        computed_tag_ids = self._get_or_create_data_tag_ids(tags)
        source = self._get_or_create_data_source(source)

        if metadata and isinstance(metadata, dict):
            metadata = [metadata]

        if custom_metadata and isinstance(custom_metadata, dict):
            custom_metadata = [custom_metadata]

        if isinstance(filepaths, str) or isinstance(filepaths, Path):
            filepaths = [filepaths]

        if metadata and len(metadata) != len(filepaths):
            raise BadRequestError(
                f"Given list of metadata has {len(metadata)} objects but list of paths has {len(filepaths)} objects."
                f"Please give the same number of objects if you want to have metadata added to your data"
            )

        if custom_metadata and len(custom_metadata) != len(filepaths):
            raise BadRequestError(
                f"Given list of custom metadata has {len(custom_metadata)} objects but list of paths has {len(filepaths)} objects."
                f"Please give the same number of objects if you want to have custom metadata added to your data"
            )

        uploader = DataUploader(
            self.connexion,
            self.id,
            self.connector_id,
            computed_tag_ids=computed_tag_ids,
            source=source,
            fill_metadata=fill_metadata,
            upload_dir=upload_dir,
            error_manager=error_manager,
        )

        def _upload(items: tuple[str | Path, dict | None, dict | None]):
            response = uploader.upload(items)
            if not response:
                return None

            return Data(self.connexion, self.id, response.json())

        logger.info("🌎 Starting upload..")

        # Create batches from filepaths and metadata
        batches = [
            (
                filepaths[k],
                metadata[k] if metadata else None,
                custom_metadata[k] if custom_metadata else None,
            )
            for k in range(len(filepaths))
        ]

        results = mlt.do_mlt_function(
            batches, _upload, h=lambda batch: batch[0], max_workers=max_workers
        )

        error_data = []
        pending_data = []
        uploaded_data = []
        for _, data in results.items():
            if data is None:
                continue

            if data.upload_status == DataUploadStatus.ERROR:
                if error_manager:
                    error_manager.append(UnprocessableData(data))
                error_data.append(data)
            else:
                uploaded_data.append(data)
                if not data.is_ready():
                    pending_data.append(data)

        if len(uploaded_data) != len(filepaths) or len(error_data) > 0:
            logger.error(
                f"❌ {len(filepaths) - len(uploaded_data) + len(error_data)} data not uploaded."
            )
            if error_manager:
                logger.error(
                    "Calling error_manager.errors will return a list of UploadError objects to see what happened"
                )

        if len(uploaded_data) == 0:
            raise NothingDoneError("Nothing has been uploaded.")
        elif len(uploaded_data) == 1:
            first = uploaded_data[0]
            if (
                first.upload_status != DataUploadStatus.DONE
                and wait_for_unprocessable_data
            ):
                logger.info(
                    f"{first.filename} data is being processed on Picsellia, please wait a few moment.."
                )
                first.wait_for_upload_done(blocking_time_increment=5.0, attempts=30)
            logger.info(f"✅ {first.filename} data uploaded in {self}")
            return first
        else:
            if wait_for_unprocessable_data and len(pending_data) > 0:
                pending_multi_data = MultiData(self.connexion, self.id, pending_data)
                pending_multi_data.wait_for_upload_done(
                    blocking_time_increment=5.0, attempts=30
                )
            logger.info(f"✅ {len(uploaded_data)} data uploaded in {self}")
            return MultiData(self.connexion, self.id, uploaded_data)

    @exception_handler
    @beartype
    def find_data(
        self,
        filename: str | None = None,
        object_name: str | None = None,
        id: str | UUID | None = None,
    ) -> Data:
        """Find a data into this datalake

        You can find it by giving its filename or its object name or its id

        Examples:
            ```python
            my_data = my_datalake.find_data(filename="test.png")
            ```
        Arguments:
            filename (str, optional): filename of the data. Defaults to None.
            object_name (str, optional): object name in the storage S3. Defaults to None.
            id (str or UUID, optional): id of the data. Defaults to None

        Raises:
            If no data match the query, it will raise a NotFoundError.
            In some case, it can raise an InvalidQueryError,
                it might be because platform stores 2 data matching this query (for example if filename is duplicated)

        Returns:
            The (Data) found
        """
        assert not (
            filename is None and object_name is None and id is None
        ), "Select at least one criteria to find a data"

        params = {}
        if id is not None:
            params["id"] = id

        if filename is not None:
            params["filename"] = filename

        if object_name is not None:
            params["object_name"] = object_name

        params = add_data_mandatory_query_parameters(params)

        r = self.connexion.get(
            f"/api/datalake/{self.id}/datas/find", params=params
        ).json()
        return Data(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def list_data(
        self,
        limit: int | None = None,
        offset: int | None = None,
        page_size: int | None = None,
        order_by: list[str] | None = None,
        tags: str | Tag | list[str | Tag] | None = None,
        filenames: list[str] | None = None,
        intersect_tags: bool | None = False,
        object_names: list[str] | None = None,
        q: str | None = None,
        ids: list[str | UUID] | None = None,
        custom_metadata: dict | None = None,
    ) -> MultiData:
        """List data of this datalake.

        If there is no data, raise a NoDataError exception.

        Returned object is a MultiData. An object that allows manipulation of a bunch of data.
        You can add tags on them or feed a dataset with them.

        Examples:
            ```python
            lake = client.get_datalake()
            data = lake.list_data()
            ```

        Arguments:
            limit (int, optional): if given, will limit the number of data returned
            offset (int, optional): if given, will return data that would have been returned
                                    after this offset in given order
            page_size (int, optional): deprecated.
            order_by (list[str], optional): if not empty, will order data by fields given in this parameter
            filenames (list[str], optional): if given, will return data that have filename equals to one of given filenames
            object_names (list[str], optional): if  given, will return data that have object name equals to one of given object names
            tags (str, (Tag), list[(Tag) or str], optional): if given, will return data that have one of given tags
                                                            by default. if `intersect_tags` is True, it will return data
                                                            that have all the given tags
            intersect_tags (bool, optional): if True, and a list of tags is given, will return data that have
                                             all the given tags. Defaults to False.
            q (str, optional): if given, will filter data with given query. Defaults to None.
            ids: (list[UUID]): ids of the data you're looking for. Defaults to None.
            custom_metadata: (dict, optional): filter based on the custom_metadata linked to the Data. Defaults to None

        Raises:
            NoDataError: When datalake has no data, raise this exception.

        Returns:
            A (MultiData) object that wraps a list of (Data).
        """
        if page_size:
            logger.warning("page_size is deprecated and not used anymore.")

        tags_q = convert_tag_list_to_query_language(tags, intersect_tags)
        query = combine_two_ql(q, tags_q)

        filters = DataFilter.model_validate(
            {
                "limit": limit,
                "offset": offset,
                "order_by": order_by,
                "filenames": filenames,
                "object_names": object_names,
                "ids": ids,
                "query": query,
                "custom_metadata": custom_metadata,
            }
        )
        data = DataLister(self.connexion, self.id).list_items(filters)

        if len(data) == 0:
            raise NoDataError("No data found in this datalake with this query")

        return MultiData(self.connexion, self.id, data)

    @exception_handler
    @beartype
    def create_data_tag(self, name: str) -> Tag:
        """Create a data tag used in this datalake

        Examples:
            ```python
            tag_car = lake.create_data_tag("car")
            ```
        Arguments:
            name (str): Name of the tag to create

        Returns:
            A (Tag) object
        """
        payload = {"name": name}
        r = self.connexion.post(
            f"/api/datalake/{self.id}/tags", data=orjson.dumps(payload)
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def get_data_tag(self, name: str) -> Tag:
        """Retrieve a data tag used in this datalake.

        Examples:
            ```python
            tag_car = lake.get_data_tag("car")
            ```

        Arguments:
            name (str): Name of the tag to retrieve

        Returns:
            A (Tag) object
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/datalake/{self.id}/tags/find", params=params
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def get_or_create_data_tag(self, name: str) -> Tag:
        """Retrieve a data tag used in this datalake by its name.
        If tag does not exist, create it and return it.

        Examples:
            ```python
            tag = lake.get_or_create_data_tag("new_tag")
            ```

        Arguments:
            name (str): Name of the tag to retrieve or create

        Returns:
            A (Tag) object
        """
        try:
            return self.get_data_tag(name)
        except exceptions.ResourceNotFoundError:
            return self.create_data_tag(name)

    @exception_handler
    @beartype
    def list_data_tags(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Tag]:
        """List all tags of this datalake

        Examples:
            ```python
            tags = lake.list_data_tags()
            assert tag_car in tags
            ```

        Arguments:
            limit (int, optional): Limit the number of tags returned. Defaults to None.
            offset (int, optional): Offset to start listing tags. Defaults to None.
            order_by (list[str], optional): Order the tags returned by given fields. Defaults to None.

        Returns:
            A List of (Tag)
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(f"/api/datalake/{self.id}/tags", params=params).json()
        return [Tag(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def create_projection(
        self,
        data: Data,
        name: str,
        path: str,
        additional_info: dict = None,
        fill_metadata: bool = False,
    ) -> DataProjection:
        """Attach a Projection to an already existing Data.
        A Projection is another file that will be viewable along the original Data in the UI and in the annotation (if the type is compatible with the web browser).
        You can add as many Projections to a Data as you want.
        The type of this projection will be set to 'CUSTOM'

        Arguments:
            data (Data): target (Data).
            name (str): projection name.
            path (str): path of the file to upload
            additional_info (dict, optional): some data to attach to your projection. Defaults to None
            fill_metadata (bool, optional): if true, we will read image and add exif metadata to your projection. Defaults to False.

        Returns:
            A (DataProjection) object
        """

        info = additional_info or {}
        image_metadata = DataUploader.prepare_metadata_from_image(
            path, info, fill_metadata
        )

        filename = os.path.basename(path)
        file_size = Path(path).stat().st_size
        object_name = self.connexion.generate_data_object_name(
            filename, ObjectDataType.DATA_PROJECTION, self.connector_id
        )
        _, _, content_type = self.connexion.upload_file(
            object_name, path, connector_id=self.connector_id
        )
        return self._create_projection_from_cloud_image(
            data,
            name,
            object_name,
            filename,
            content_type,
            file_size,
            image_metadata.width,
            image_metadata.height,
            image_metadata.metadata,
        )

    @exception_handler
    @beartype
    def _create_projection_from_cloud_image(
        self,
        data: Data,
        name: str,
        object_name: str,
        filename: str,
        content_type: str,
        file_size: int,
        width: int,
        height: int,
        metadata: dict | None,
    ) -> DataProjection:
        payload = {
            "name": name,
            "file_size": file_size,
            "object_name": object_name,
            "filename": filename,
            "content_type": content_type,
            "width": width,
            "height": height,
            "additional_info": metadata,
        }
        item = self.connexion.post(
            f"/api/data/{data.id}/projections",
            data=orjson.dumps(payload),
        ).json()
        logger.info(f"Projection {name} created for data {data.id}")
        return DataProjection(self.connexion, data.id, item)

    @exception_handler
    @beartype
    def import_bucket_objects(
        self,
        prefixes: list[str],
        tags: list[str | Tag] | None = None,
        source: str | DataSource | None = None,
    ) -> Job:
        """Asynchronously import files from your remote storage (bucket) into this (Datalake)
        Only files with known content-types will be added.

        This method takes a list of prefixes. Prefixes can either be full object names or the prefix to a bunch of object names
        Given tags and source will be added to all imported data.
        We will read exif of your image to create metadata.

        You can only call this method if you use a private object storage with this datalake, owned by your organization.
        You should use this method carefully as it can import your whole S3 in the platform if you import for example "/"

        If you want to import projections from your object storage, or if you want to add custom_metadata,
        you could instead call import_cloud_objects()

        This method will return a Job object, you can call job.wait_for_done() to wait for import.
        As it might be a long task, we don't wait_for_done() in the method.

        Args:
            prefixes: list of prefixes to import
            tags: list of tags that will be added to data
            source: data source that will be specified on data

        Returns:
            A (Job) that you can wait for done.
        """

        tag_ids = self._get_or_create_data_tag_ids(tags)
        source = self._get_or_create_data_source(source)

        payload = {
            "datalake_id": self.id,
            "prefixes": prefixes,
            "tag_ids": tag_ids,
        }
        if source:
            payload["data_source_id"] = source.id

        return self._request_import_bucket_objects(payload)

    @exception_handler
    @beartype
    def import_cloud_objects(self, cloud_objects: dict[str, dict | CloudObject]) -> Job:
        """Asynchronously import files from your bucket into this Datalake.
        Only files with known content-types will be added.

        This method is limited to 500 elements.
        If you have more elements to import, consider batching and calling this method multiple times.

        The keys are the object_names of the data you want to import, the values are CloudObjects that represents all the additional information that needs to be stored against your Data.
        CloudObject is defined in picsellia.types.schemas, it's a pydantic model, but you can also give a dict, SDK will try to parse it.

        CloudObject allows: \n
        - metadata (dict, optional): metadata to attach to the data that will be created on the platform. It must match requirements from https://documentation.picsellia.com/update/docs/datalake-metadata#/
        - custom_metadata (dict): a dict of metadata to attach to the created Data
        - tags (list[str]): a list of tag names, the SDK will get or create each name
        - data_source (str): a str, the SDK will get or create source

        You can only call this method if you use a private object storage with this datalake, owned by your organization.

        This will launch one asynchronous job, it is returned by this method and can be waited

        Examples:
            ```python
                from picsellia.types.schemas import CloudObject, CloudProjectionObject
                datalake = client.get_datalake()
                job = datalake.import_cloud_objects(
                    cloud_objects={
                        "/bucket/path/object-1.jpg": {
                            "metadata": {
                                "reference": "XYZ1"
                            },
                            "tags": ["tag-1", "tag-2"],
                            "data_source": "cloud",
                            "custom_metadata": {"value": 10},
                        },
                        "/bucket/path/object-2.jpg": CloudObject(
                            "metadata"={
                                "reference": "XYZ1"
                            },
                            tags=["tag-1"],
                            data_source="cloud",
                            custom_metadata={"value": 25},
                        ),
                    }
                )
                job.wait_for_done()
                datalake.import_cloud_projections(
                    cloud_projections={
                        "/bucket/path/object-1.jpg": [
                            {
                                "name": "view",
                                "object_name": "/bucket/path/object-1-projection.jpg",
                            }
                        ],
                        "/bucket/path/object-2.jpg": [
                                CloudProjectionObject(
                                    name="pr1",
                                    object_name="/bucket/path/object-2-projection.jpg",
                                )
                        ],
                    }
                )
            ```
        Args:
            cloud_objects (dict): dict with object names as keys and CloudObject as values
        Returns:
            A (Job) that you can wait for done.
        """
        if len(cloud_objects) < 1 or len(cloud_objects) > 500:
            raise ValueError(
                "Parameter 'cloud_objects' must have at least 1 and less than 500 elements."
            )

        payload = {"datalake_id": self.id, "objects": {}}
        for name, obj in cloud_objects.items():
            payload["objects"][name] = (
                CloudObject(**obj).model_dump()
                if isinstance(obj, dict)
                else obj.model_dump()
            )

        return self._request_import_bucket_objects(payload)

    def _request_import_bucket_objects(self, payload: dict) -> Job:
        r = self.connexion.post(
            path=f"/api/datalake/{self.id}/objects",
            data=orjson.dumps(payload),
        ).json()
        self.refresh(r["datalake"])
        return Job(self.connexion, r["job"], version=2)

    @exception_handler
    @beartype
    def import_cloud_projections(
        self, cloud_projections: dict[str, list[dict | CloudProjectionObject]]
    ) -> Job:
        """Asynchronously import files from your bucket as (DataProjection) into this Datalake.
        Only files with known content-types will be added.

        This method is limited to 500 elements.
        If you have more elements to import, consider batching and calling this method multiple times.

        The keys must be object names of Data that are ALREADY exist in your Datalake. The values are a list of CloudProjectionObjects (or the corresponding dict), that represents a (DataProjection).
        CloudProjectionObject is defined in picsellia.types.schemas, it's a pydantic model, but you can also give a dict, SDK will try to parse it.

        CloudProjectionObject must have: \n
        - name (str): name of your projection
        - object_name (str): path in your bucket of your projection file.

        You can only call this method if you use a private object storage with this datalake, owned by your organization.

        This will launch one asynchronous job, it is returned by this method and can be waited

        Examples:
            ```python
                from picsellia.types.schemas import CloudObject, CloudProjectionObject
                datalake = client.get_datalake()
                datalake.import_cloud_projections(
                    cloud_projections={
                        "/bucket/path/object-1.jpg": [
                            {
                                "name": "view",
                                "object_name": "/bucket/path/object-1-projection.jpg",
                            }
                        ],
                        "/bucket/path/object-2.jpg": [
                            CloudProjectionObject(
                                name="pr1",
                                object_name="/bucket/path/object-2-projection.jpg",
                            )
                        ],
                    }
                )
            ```
        Args:
            cloud_projections (dict): dict with object names as keys and CloudProjectionObject as values
        Returns:
            A (Job) that you can wait for done.
        """
        if len(cloud_projections) < 1 or len(cloud_projections) > 500:
            raise ValueError(
                "Parameter 'cloud_projections' must have at least 1 and less than 500 elements."
            )

        payload = {"datalake_id": self.id, "projections": {}}
        for name, projections in cloud_projections.items():
            payload["projections"][name] = [
                (
                    CloudProjectionObject(**obj).model_dump()
                    if isinstance(obj, dict)
                    else obj.model_dump()
                )
                for obj in projections
            ]

        r = self.connexion.post(
            path=f"/api/datalake/{self.id}/objects/projections",
            data=orjson.dumps(payload),
        ).json()
        self.refresh(r["datalake"])
        return Job(self.connexion, r["job"], version=2)

    def _get_or_create_data_tag_ids(self, tags: list[str | Tag] | None):
        tag_ids = []
        if tags:
            for tag in tags:
                if isinstance(tag, str):
                    computed_tag = self.get_or_create_data_tag(tag)
                else:
                    if tag.target_type == TagTarget.DATA:
                        computed_tag = tag
                    else:
                        computed_tag = self.create_data_tag(tag.name)

                tag_ids.append(computed_tag.id)
        return tag_ids

    def _get_or_create_data_source(self, source: str | DataSource | None):
        if isinstance(source, str):
            return DataSourceService.get_or_create_datasource(
                self.connexion, self._organization_id, source
            )
        else:
            return source

    @exception_handler
    @beartype
    def launch_processing(
        self,
        processing: Processing,
        data: list[Data] | MultiData,
        parameters: dict = None,
        cpu: int = None,
        gpu: int = None,
        model_version_id: UUID = None,
        target_datalake_name: str = None,
    ) -> Job:
        """Launch given processing onto this datalake version. You can give specific cpu, gpu or parameters.
        You can give a model_version_id used by the processing. Constraints defined by the processing will be checked before launching.
        You can give a target_datalake_name, it will create a Datalake, and the processing will be able to use this as output_datalake

        If not given, it will use default values specified in Processing.
        If processing cannot be launched on this Datalake it will raise before launching.

        Examples:
            ```python
            processing = client.get_processing("data auto tagging")
            data = datalake.list_data(limit=10)
            datalake.launch_processing(processing, data)
            ```

        Returns:
            A (Job) object
        """
        data_ids = list({datum.id for datum in data})
        payload = {
            "processing_id": processing.id,
            "parameters": parameters,
            "data_ids": data_ids,
            "cpu": cpu,
            "gpu": gpu,
            "model_version_id": model_version_id,
            "target_datalake_name": target_datalake_name,
        }
        r = self.connexion.post(
            f"/api/datalake/{self.id}/processing/launch",
            data=orjson.dumps(payload),
        ).json()
        return Job(self.connexion, r, version=2)

    @exception_handler
    @beartype
    def embeddings_computation_status(self):
        """Return the status of the Visual Search for this (Datalake)

        Returns:
            a dict with status
        """
        return self.connexion.get(
            f"/api/visual-search/datalake/{self.id}/status"
        ).json()

    @exception_handler
    @beartype
    def visual_search(self, data: Data, limit: int) -> MultiData:
        """
        Will return a (MultiData) object with data that are similar to given (Data), ordered by score.
        This is computed with Visual Search feature.
        Each data will have a temporary attribute _score, if you want to access the similarity score.

        Returns:
            a MultiData object
        """
        return DatalakeVisualSearchService(self.connexion, self.id).query_visual_search(
            str(data.id), limit
        )

    @exception_handler
    @beartype
    def text_search(self, query: str, limit: int) -> MultiData:
        """
        Will return a MultiData object with data that match your query, ordered by score.
        This is computed with Visual Search feature.
        Each data will have a temporary attribute _score, if you want to access the similarity score.

        Returns:
            a MultiData object
        """
        return DatalakeVisualSearchService(self.connexion, self.id).query_visual_search(
            query, limit
        )

    @exception_handler
    @beartype
    def count_embeddings(self) -> int:
        """Return the number of data indexed by the Visual Search in this (Datalake)

        Returns:
            number of data indexed
        """
        r = self.connexion.get(
            f"/api/visual-search/datalake/{self.id}/points/count"
        ).json()
        return r["count"]

    @exception_handler
    @beartype
    def list_embeddings(self, limit: int) -> list[dict]:
        """Return the list of embeddings computed for this (Datalake)

        Returns:
            a list of dict with data of indexation
            each dictionary contains:
                - id (str): UUID of the (Data)
                - vector (dict): Model-specific vector embeddings where:
                    - key (str): Embedder identifier
                    - value (list): Vector embedding as list of floats
        """
        return DatalakeVisualSearchService(self.connexion, self.id).list_embeddings(
            limit, with_vector=True, with_payload=False, has_error=False
        )



import logging
import os
from uuid import UUID

import orjson
import semantic_version
from beartype import beartype
from requests import Session

import picsellia
from picsellia import utils
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import PicselliaError
from picsellia.sdk.annotation_campaign import AnnotationCampaign
from picsellia.sdk.connexion import Connection
from picsellia.sdk.datalake import Datalake
from picsellia.sdk.dataset import Dataset
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.deployment import Deployment
from picsellia.sdk.experiment import Experiment
from picsellia.sdk.job import Job
from picsellia.sdk.model import Model
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.processing import Processing
from picsellia.sdk.project import Project
from picsellia.sdk.review_campaign import ReviewCampaign
from picsellia.sdk.tag import Tag
from picsellia.sdk.user import User
from picsellia.services.datasource import DataSourceService
from picsellia.types.enums import InferenceType, ProcessingType, TagTarget
from picsellia.types.schemas import OrganizationSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Client:
    """Picsellia SDK Client can be used to communicate with Picsellia services.
    You need an API Token, available on web platform in Profile Settings > Token.
    From this client object, you will be able to retrieve python objects representing your data in the platform.

    Examples:
        ```python
        client = Client(api_token="a0c9f3bbf7e8bc175494fc44bfc6f89aae3eb9d0")
        ```

    Arguments:
        api_token (str, optional): your api token, that can be found on the web platform.
            Defaults to None, client will try to find it into your environment variable as 'PICSELLIA_TOKEN'
        organization_id: (str or UUID, optional): Specify an organization to connect to by giving its id.
            Defaults to None, you will be connected to your main Organization.
        organization_name: (str, optional): Specify an organization to connect to by giving its name.
            If id is also given, client will use organization_id.
            Defaults to None, you will be connected to your main Organization.
        host (str, optional): Define a custom host used for platform.
            Defaults to our Picsellia environment "https://app.picsellia.com".
        session (requests.Session, optional): Set up your own requests.Session object to add headers or proxies.
    """

    def __init__(
        self,
        api_token: str | None = None,
        organization_id: str | UUID | None = None,
        organization_name: str | None = None,
        host: str = "https://app.picsellia.com",
        session: Session | None = None,
    ):
        if api_token is None:
            if "PICSELLIA_TOKEN" in os.environ:
                token = os.environ["PICSELLIA_TOKEN"]
            else:
                raise PicselliaError(
                    "Please set up the PICSELLIA_TOKEN environment variable or specify your token"
                )
        else:
            token = api_token

        self.connexion = Connection(host, token, session=session)

        # Ping platform to get username and version matching api_token
        try:
            ping_response = self.connexion.get("/api/home/ping").json()
        except Exception as e:  # pragma: no cover
            raise PicselliaError(
                "Cannot connect to the platform. Please check api_token, organization and host given.\n"
                f"Error is : {e}"
            )

        sdk_version = picsellia.__version__
        platform_version = ping_response["sdk_version"]
        if self.is_sdk_outdated(sdk_version, platform_version):  # pragma: no cover
            logger.warning(
                f"\033[93mYou are using an outdated version of the picsellia package ({sdk_version})\033[0m"
            )
            logger.warning(
                f"\033[93mPlease consider upgrading to {platform_version} with pip install picsellia --upgrade\033[0m"
            )

        if organization_id is not None:
            if isinstance(organization_id, str):
                organization_id = UUID(organization_id)
            self._id = organization_id
            self.sync()
        elif organization_name is not None:
            r = self.connexion.get(
                path="/api/organizations/find", params={"name": organization_name}
            ).json()
            self._id = r["id"]
            self.sync()
        elif (
            "default_organization" in ping_response
            and ping_response["default_organization"]
        ):
            self._id = ping_response["default_organization"]
            self.sync()
        else:
            raise PicselliaError(
                "You cannot create a Client with no organization selected when you don't own any organization."
            )

        if self._id == ping_response["default_organization"]:
            message = "your"
        else:  # pragma: no cover
            message = self.name + "'s"

        username = ping_response["username"]
        logger.info(f"Hi {Colors.BLUE}{username}{Colors.ENDC}, welcome back. 🥑")
        logger.info(f"Workspace: {Colors.YELLOW}{message}{Colors.ENDC} organization.")

    def __str__(self) -> str:
        return f"Client initialized for organization `{self.name}`"

    def __enter__(self):
        self.connexion = self.connexion.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connexion.__exit__(exc_type, exc_val, exc_tb)

    @property
    def id(self) -> UUID:
        """Organization UUID connected with this (Client)"""
        return self._id

    @property
    def name(self) -> str:
        """Organization name connected with this (Client)"""
        return self._name

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/organization/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> OrganizationSchema:
        schema = OrganizationSchema(**data)
        self._name = schema.name
        if schema.default_connector_id is not None:
            self.connexion.connector_id = schema.default_connector_id
            self.connexion.organization_id = schema.id
        return schema

    @classmethod
    def is_sdk_outdated(cls, sdk: str, platform: str):
        try:
            sdk_version = semantic_version.Version(sdk)
            platform_version = semantic_version.Version(platform)
        except ValueError:
            return False

        return sdk_version < platform_version

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
            print(foo_dataset.get_resource_url_on_platform())
            >>> "https://app.picsellia.com/organization/62cffb84-b92c-450c-bc37-8c4dd4d0f590"
            ```

        Returns:
            Url on Platform for this resource
        """

        return f"{self.connexion.host}/organization/{self.id}"

    @exception_handler
    @beartype
    def list_users(self, username: str | None = None) -> list[User]:
        """Retrieve all users of this organization
        Returns:
            a list of (User)
        """
        params = {}
        if username:
            params["username"] = username
        r = self.connexion.get(
            f"/api/access/organization/{self.id}/users", params=params
        ).json()
        return [User(self.connexion, item) for item in r["users"]]

    @exception_handler
    @beartype
    def get_datalake(
        self, id: str | UUID | None = None, name: str | None = None
    ) -> Datalake:
        """Retrieve a datalake of your organization.
        By default, is nothing is given, this will return your default datalake.

        Examples:
            ```python
            datalake = client.get_datalake()
            datalake = client.get_datalake(name="default")
            datalake = client.get_datalake(id="0188e773-db65-7546-8e3e-b38fe5bed6d2")
            ```

        Arguments:
            id (str or UUID, optional): id of the datalake to retrieve. Defaults to None.
            name (str, optional): name of the datalake to retrieve. Defaults to None.

        Returns:
            The (Datalake) of the client that you are using
        """
        params = {}
        if id:
            params["id"] = UUID(id) if isinstance(id, str) else id
        elif name:
            params["name"] = name

        r = self.connexion.get(
            f"/api/organization/{self.id}/datalake", params=params
        ).json()

        return Datalake(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def list_datalakes(self) -> list[Datalake]:
        """Retrieve all datalakes linked to this organization

        Examples:
            ```python
            datalakes = client.list_datalakes()
            ```

        Returns:
            List of (Datalake) of the client that you are using
        """
        # Retrieve default datalake
        r = self.connexion.get(f"/api/organization/{self.id}/datalakes").json()
        return [Datalake(self.connexion, self.id, item) for item in r["items"]]

    @exception_handler
    @beartype
    def create_dataset(
        self,
        name: str,
        description: str = "",
        private: bool = None,
    ) -> Dataset:
        """Create a (Dataset) in this organization.

        This method allows user to create a dataset into the organization currently connected.
        A dataset can then be versioned into (DatasetVersion).
        User can specify name of the dataset, a description and if the dataset is private or not.

        Examples:
            Create a dataset named datatest with data from datalake and version it
            ```python
            foo_dataset = client.create_dataset('foo_dataset')
            foo_dataset_version_1 = foo_dataset.create_version('first')
            some_data = client.get_datalake().list_data(limit=10)
            foo_dataset_version_1.add_data(some_data)
            ```

        Arguments:
            name (str): Name of the dataset. It must be unique in the organization.
            description (str, optional): A description of the dataset. Defaults to ''.
            private (bool, optional): Specify if the dataset is private. Defaults to True.

        Returns:
            A (Dataset) that you can manipulate, connected to Picsellia
        """
        assert name != "", "Dataset name can't be empty"

        if private is not None:
            logging.warning(
                "'private' parameter is deprecated and will be removed in future versions. "
                "You cannot create a public dataset from the SDK anymore."
            )

        payload = {"name": name, "description": description}
        r = self.connexion.post(
            f"/api/organization/{self.id}/datasets",
            data=orjson.dumps(payload),
        ).json()
        return Dataset(self.connexion, r)

    @exception_handler
    @beartype
    def get_dataset(self, name: str) -> Dataset:
        """Retrieve a dataset by its name

        Examples:
            ```python
            foo_dataset = client.get_dataset('datatest')
            foo_dataset_version = foo_dataset.get_version('first')
            ```

        Arguments:
            name (str): Name of the dataset

        Returns:
            A (Dataset) that you can use and manipulate
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/organization/{self.id}/datasets/find",
            params=params,
        ).json()
        return Dataset(self.connexion, r)

    @exception_handler
    @beartype
    def get_dataset_by_id(self, id: UUID | str) -> Dataset:
        """Get a dataset by its id

        Examples:
            ```python
            dataset = client.get_dataset_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str): id of the dataset to retrieve

        Returns:
            A (Dataset) that you can use and manipulate
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/datasets/find", params=params
        ).json()
        return Dataset(self.connexion, r)

    @exception_handler
    @beartype
    def list_datasets(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Dataset]:
        """Retrieve all dataset of current organization

        Examples:
            ```python
            datasets = client.list_datasets()
            ```

        Arguments:
            limit (int, optional): Limit number of datasets to retrieve. Defaults to None, all datasets will be retrieved.
            offset (int, optional): Offset to begin with when listing datasets. Defaults to None, starting at 0.
            order_by (list[str], optional): Some fields to order datasets against. Defaults to None, datasets will not be sorted

        Returns:
            A list of (Dataset) object that belongs to your organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/api/organization/{self.id}/datasets", params=params
        ).json()
        return [Dataset(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def get_dataset_version_by_id(self, id: UUID | str) -> DatasetVersion:
        """Get a dataset version by its id

        Examples:
            ```python
            dataset_version = client.get_dataset_version_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str or UUID): id of the dataset version to retrieve

        Returns:
            A (DatasetVersion)
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/datasetversions/find", params=params
        ).json()
        return DatasetVersion(self.connexion, r)

    @exception_handler
    @beartype
    def create_model(self, name: str, description: str = None) -> Model:
        """Create a new model.

        Examples:
            ```python
            model = client.create_model(name="foo_model", description="A brand new model!")
            ```

        Arguments:
            name (str): Model name to create.
            description (str, optional): Description of this model. Defaults to None

        Returns:
            A (Model) object that you can manipulate
        """
        if description is None:
            description = "A brand new model!"

        payload = {
            "name": name,
            "description": description,
        }
        r = self.connexion.post(
            f"/api/organization/{self.id}/models",
            data=orjson.dumps(payload),
        ).json()
        created_model = Model(self.connexion, r)
        logger.info(
            f"📚 Model {created_model.name} created\n📊 🌐 Platform url: {created_model.get_resource_url_on_platform()}"
        )
        return created_model

    @exception_handler
    @beartype
    def get_model_by_id(self, id: UUID | str) -> Model:
        """Retrieve a model by its id

        Examples:
            ```python
            model = client.get_model_by_id(UUID("d8fae655-5c34-4a0a-a59a-e49c89f20998"))
            ```
        Arguments:
            id (str): id of the model that you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/models/find", params=params
        ).json()
        return Model(self.connexion, r)

    def _do_get_model_by_name(self, url: str, name: str):
        params = {"name": name}
        r = self.connexion.get(url, params=params).json()
        return Model(self.connexion, r)

    @exception_handler
    @beartype
    def get_model(self, name: str) -> Model:
        """Retrieve a model by its name.

        Examples:
            ```python
            model = client.get_model("foo_model")
            ```
        Arguments:
            name (str): name of the model you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        return self._do_get_model_by_name(
            f"/api/organization/{self.id}/models/find", name
        )

    @exception_handler
    @beartype
    def get_public_model(self, name: str) -> Model:
        """Retrieve a public model by its name.
           It can only retrieve *public* model.

        Examples:
            ```python
            model = client.get_public_model("foo_public_model")
            ```
        Arguments:
            name (str): name of the public model you are looking for

        Returns:
            A (Model) object that you can manipulate
        """
        return self._do_get_model_by_name("/api/models/find", name)

    def _do_list_models(self, url: str, params: dict):
        params = filter_payload(params)
        r = self.connexion.get(url, params=params).json()
        return [Model(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def list_models(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Model]:
        """List all models stored in this organization

        This will return all the models stored
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
            models = client.list_models()
            ```

        Arguments:
            limit (int, optional): Limit number of models to retrieve. Defaults to None, all models will be retrieved.
            offset (int, optional): Offset to begin with when listing models. Defaults to None, starting at 0.
            order_by (list[str], optional): Some fields to order models against. Defaults to None, models will not be sorted

        Returns:
            A list of all (Model) that belong to this organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        return self._do_list_models(
            f"/api/organization/{self.id}/models", params=params
        )

    @exception_handler
    @beartype
    def list_public_models(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
        name: str | None = None,
        type: str | InferenceType | None = None,
    ) -> list[Model]:
        """List all public models of Picsellia Hub
        This will return all public models of the hub

        Arguments:
            limit (int, optional): Limit number of public models to retrieve.
                Defaults to None, all public models will be retrieved.
            offset (int, optional): Offset to begin with when listing public models.
                Defaults to None, starting at 0.
            order_by (list[str], optional): Some fields to order models against.
                Defaults to None, models will not be sorted
            name (str, optional): A name to filter public models. It will return models with name *containing*
                this parameter. Defaults to None, models will not be filtered
            type (str or InferenceType, optional): A type to filter public models on.

        Examples:
            ```python
            public_detection_models = client.list_public_models(name="yolo")
            ```

        Returns:
            A list of all public (Model) objects
        """
        if type:
            logging.warning(
                "'type' parameter is deprecated and will be removed in future versions."
            )

        params = {
            "limit": limit,
            "offset": offset,
            "order_by": order_by,
            "name": name,
        }
        return self._do_list_models("/api/models", params=params)

    @exception_handler
    @beartype
    def get_model_version_by_id(self, id: UUID | str) -> ModelVersion:
        """Get a model version by its id

        Examples:
            ```python
            model_version = client.get_model_version_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str or UUID): id of the model version to retrieve

        Returns:
            A (ModelVersion)
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/modelversions/find", params=params
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def create_project(
        self,
        name: str,
        description: str | None = None,
        private: bool | None = True,
    ) -> Project:
        """Create a project with given name and parameters

        This project will be registered into used organization.

        Examples:
            ```python
            my_project = client.create_project("my_project", description="My first project!")
            ```
        Arguments:
            name (str): name of the project
            description (str): description of the project
            private (bool, deprecated): This parameter is deprecated

        Returns:
            A (Project) that you can manipulate to run experiments, or attach dataset
        """
        payload = {"name": name}

        if private is False:
            logger.warning(
                "You cannot create a public project anymore. This parameter will not be used"
            )

        if description is not None:
            payload["description"] = description

        r = self.connexion.post(
            f"/api/organization/{self.id}/projects",
            data=orjson.dumps(payload),
        ).json()
        created_project = Project(self.connexion, r)
        logger.info(
            f"📚 Project {name} created\n📊 Description: {description if description is not None else ''}\n🌐 Platform url: {created_project.get_resource_url_on_platform()}"
        )
        return created_project

    @exception_handler
    @beartype
    def get_project(self, project_name: str) -> Project:
        """Get a project from its name

        Retrieve a project from its name.
        Project must belong to used organization.

        Examples:
            ```python
            my_project = client.get_project("my_project")
            ```
        Arguments:
            project_name (str): name of the project to retrieve

        Returns:
            A (Project) of your organization, you can manipulate to run experiments, or attach dataset
        """
        r = self.connexion.get(
            f"/api/organization/{self.id}/projects/find",
            params={"name": project_name},
        ).json()
        return Project(self.connexion, r)

    @exception_handler
    @beartype
    def get_project_by_id(self, id: UUID | str) -> Project:
        """Get a project from its id

        Retrieve a project from its id.
        Project must belong to used organization.
        If no project is found, will throw a ResourceNotFoundError

        Examples:
            ```python
            my_project = client.get_project_by_id("2214aacc-b884-41e1-b70f-420c0cd7eefb")
            ```
        Arguments:
            id (str): id of the project to retrieve

        Returns:
            A (Project) of your organization, you can manipulate to run experiments, or attach dataset
        """
        if isinstance(id, str):
            id = UUID(id)
        r = self.connexion.get(
            f"/api/organization/{self.id}/projects/find?id={id}"
        ).json()
        return Project(self.connexion, r)

    @exception_handler
    @beartype
    def list_projects(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Project]:
        """List all projects of your organization.

        Retrieve all projects of your organization

        Examples:
            ```python
            projects = client.list_projects()
            ```

        Arguments:
            limit (int, optional): Limit number of projects to retrieve. Defaults to None, all projects will be retrieved.
            offset (int, optional): Offset to begin with when listing projects. Defaults to None, starting at 0.
            order_by (list[str], optional): Some fields to order projects against. Defaults to None, projects will not be sorted

        Returns:
            A list of Project of your organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/api/organization/{self.id}/projects", params=params
        ).json()
        return [Project(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def get_experiment_by_id(self, id: UUID | str) -> Experiment:
        """Get an experiment by its id

        Examples:
            ```python
            experiment = client.get_experiment_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            ```

        Arguments:
            id (str or UUID): id of the experiment to retrieve

        Returns:
            A (Experiment)
        """
        if isinstance(id, str):
            id = UUID(id)

        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/experiments/find", params=params
        ).json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def get_deployment(self, name: str) -> Deployment:
        """Get a (Deployment) from its name.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            ```
        Arguments:
            name (str): auto-generated name of your deployment.

        Returns:
            A (Deployment) object connected and authenticated to all the services.
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/organization/{self.id}/deployments/find", params=params
        ).json()
        return Deployment(self.connexion, r)

    @exception_handler
    @beartype
    def get_deployment_by_id(self, id: UUID | str) -> Deployment:
        """Get a (Deployment) from its name.

        Examples:
            ```python
            deployment = client.get_deployment_by_id(
                id="YOUR DEPLOYMENT ID"
            )
            ```
        Arguments:
            id (str): deployment id displayed in your deployment settings.

        Returns:
            A (Deployment) object connected and authenticated to all the services.
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/deployments/find", params=params
        ).json()
        return Deployment(self.connexion, r)

    @exception_handler
    @beartype
    def list_deployments(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Deployment]:
        """List all (Deployment) of your organization

        Examples:
            ```python
            our_deployments = client.list_deployments()
            ```
        Arguments:
            limit (int, optional): number max of results to return
            offset (int, optional): offset of page for pagination
            order_by (list[str], optional): keys on which deployments shall be sorted

        Returns:
            List of (Deployment): all deployments object connected and authenticated to all the services.
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/api/organization/{self.id}/deployments", params=params
        ).json()
        return [Deployment(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def create_deployment(
        self,
        model_version: ModelVersion,
        shadow_model_version: ModelVersion | None = None,
        name: str | None = None,
        min_threshold: float | None = None,
        target_datalake: Datalake | None = None,
        disable_serving: bool | None = False,
    ):
        """Create a (Deployment) from a model.

        This method allows you to create a (Deployment) on Picsellia. You will then have
        access to the monitoring dashboard and eventually a hosted endpoint.

        Examples:
            Create a serverless deployment
            ```python
            model_version = client.get_model_version_by_id('918351d2-3e96-4970-bb3b-420f33ded895')
            deployment = client.create_deployment(model_version=model_version)
            ```

        Arguments:
            model_version (ModelVersion): ModelVersion to deploy.
            shadow_model_version (ModelVersion, optional): ModelVersion to perform shadow predictions.
            name (str, optional): Name of your deployment. Defaults to a random name.
            min_threshold (float, optional): Threshold of detection scores used by models when predicting. Defaults to 0.
            target_datalake (Datalake, optional): Datalake to use when data are pushed into Picsellia.
                Defaults to organization default Datalake.
            disable_serving (bool, optional): Whether to not use Picsellia Serving. Defaults to False.


        Returns:
            A (Deployment)
        """
        from picsellia.sdk.deployment import Deployment

        payload = {
            "model_version_id": model_version.id,
            "name": name,
            "min_threshold": min_threshold,
        }
        if target_datalake is not None:
            payload["target_datalake_id"] = target_datalake.id
        if shadow_model_version:
            payload["shadow_model_version_id"] = shadow_model_version.id
        if disable_serving:
            payload["disable_serving"] = disable_serving

        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.post(
            f"/api/organization/{self.id}/deployments",
            data=orjson.dumps(filtered_payload),
        ).json()
        deployment = Deployment(self.connexion, r)
        logger.info(f"{model_version} is deployed on {deployment}")
        return deployment

    @exception_handler
    @beartype
    def create_datasource(self, name: str) -> DataSource:
        """Create a data source into this organization

        Examples:
            ```python
            data_source = client.create_datasource()
            ```

        Arguments:
            name (str): Name of the datasource to create

        Returns:
            A  (DataSource) object that belongs to your organization
        """
        return DataSourceService.create_datasource(self.connexion, self._id, name)

    @exception_handler
    @beartype
    def get_datasource(self, name: str) -> DataSource:
        """Retrieve a datasource by its name

        Examples:
            ```python
            data_sources = client.get_datasource("datasource_1")
            ```

        Arguments:
            name (str): Name of the datasource to retrieve

        Returns:
            A (DataSource) object that belongs to your organization
        """
        return DataSourceService.get_datasource(self.connexion, self._id, name)

    @exception_handler
    @beartype
    def get_or_create_datasource(self, name: str) -> DataSource:
        """Retrieve a datasource by its name or create it if it does not exist.

        Examples:
            ```python
            datasource = client.get_or_create_datasource("new_source")
            ```

        Arguments:
            name (str): Datasource name to retrieve or create

        Returns:
            A (DataSource) object
        """
        return DataSourceService.get_or_create_datasource(
            self.connexion, self._id, name
        )

    @exception_handler
    @beartype
    def list_datasources(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[DataSource]:
        """Retrieve all data source of current organization

        Examples:
            ```python
            data_sources = client.list_datasources()
            ```

        Arguments:
            limit (int, optional): Limit number of data sources to retrieve. Defaults to None, all data sources will be retrieved.
            offset (int, optional): Offset to begin with when listing data sources. Defaults to None, starting at 0.
            order_by (list[str], optional): Some fields to order data sources against. Defaults to None, data sources will not be sorted

        Returns:
            A list of (DataSource) object that belongs to your organization
        """
        return DataSourceService.list_datasources(
            self.connexion, self._id, limit, offset, order_by
        )

    @exception_handler
    @beartype
    def _create_tag_organization_scoped(self, name: str, target: TagTarget) -> Tag:
        payload = {"name": name, "target_type": target}
        r = self.connexion.post(
            f"/api/organization/{self.id}/tags",
            data=orjson.dumps(payload),
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def create_dataset_tag(self, name: str) -> Tag:
        """Create a Dataset Tag, usable only on Dataset objects

        Examples:
            ```python
            tag = client.create_dataset_tag("global")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.DATASET)

    @exception_handler
    @beartype
    def create_dataset_version_tag(self, name: str) -> Tag:
        """Create a Dataset Version Tag, usable only on Dataset Version objects

        Examples:
            ```python
            tag = client.create_dataset_version_tag("train")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.DATASET_VERSION)

    @exception_handler
    @beartype
    def create_model_tag(self, name: str) -> Tag:
        """Create a Model Tag, usable only on Model objects

        Examples:
            ```python
            tag = client.create_model_tag("model")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.MODEL)

    @exception_handler
    @beartype
    def create_model_version_tag(self, name: str) -> Tag:
        """Create a Model Version Tag, usable only on Model Version objects

        Examples:
            ```python
            tag = client.create_model_version_tag("initial")
            ```
        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.MODEL_VERSION)

    @exception_handler
    @beartype
    def create_deployment_tag(self, name: str) -> Tag:
        """Create a Deployment Tag, usable only on Deployment objects

        Examples:
            ```python
            tag = client.create_deployment_tag("operation")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._create_tag_organization_scoped(name, TagTarget.DEPLOYMENT)

    @exception_handler
    @beartype
    def _list_tags_organization_scoped(self, target: TagTarget) -> list[Tag]:
        params = {"target_type": target.value}
        r = self.connexion.get(
            f"/api/organization/{self.id}/tags", params=params
        ).json()
        return [Tag(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def list_dataset_tags(self) -> list[Tag]:
        """List all Dataset tags, usable only on Dataset objects

        Examples:
            ```python
            tags = client.list_dataset_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.DATASET)

    @exception_handler
    @beartype
    def list_dataset_version_tags(self) -> list[Tag]:
        """List all Dataset Version tags, usable only on Dataset Version objects

        Examples:
            ```python
            tags = client.list_dataset_version_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.DATASET_VERSION)

    @exception_handler
    @beartype
    def list_model_tags(self) -> list[Tag]:
        """List all Model tags, usable only on Model objects

        Examples:
            ```python
            tags = client.list_model_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.MODEL)

    @exception_handler
    @beartype
    def list_model_version_tags(self) -> list[Tag]:
        """List all Model Version tags, usable only on Model Version objects

        Examples:
            ```python
            tags = client.list_model_version_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.MODEL_VERSION)

    @exception_handler
    @beartype
    def list_deployment_tags(self) -> list[Tag]:
        """List all Deployment tags, usable only on Deployment Version objects

        Examples:
            ```python
            tags = client.list_deployment_tags()
            ```

        Returns:
            A list of (Tag) objects
        """
        return self._list_tags_organization_scoped(TagTarget.DEPLOYMENT)

    @exception_handler
    @beartype
    def _find_tag_organization_scoped(self, target: TagTarget, name: str) -> Tag:
        params = {"name": name, "target_type": target.value}
        r = self.connexion.get(
            f"/api/organization/{self.id}/tags/find",
            params=params,
        ).json()
        return Tag(self.connexion, r)

    @exception_handler
    @beartype
    def find_dataset_tag(self, name: str) -> Tag:
        """Find a Dataset tag, usable only on Dataset objects, from its name

        Examples:
            ```python
            tag = client.find_dataset_tag("global")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.DATASET, name)

    @exception_handler
    @beartype
    def find_dataset_version_tag(self, name: str) -> Tag:
        """Find a Dataset Version tag, usable only on Dataset Version objects, from its name

        Examples:
            ```python
            tag = client.find_dataset_version_tag("train")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.DATASET_VERSION, name)

    @exception_handler
    @beartype
    def find_model_tag(self, name: str) -> Tag:
        """Find a Model tag, usable only on Model objects, from its name

        Examples:
            ```python
            tag = client.find_model_tag("model")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.MODEL, name)

    @exception_handler
    @beartype
    def find_model_version_tag(self, name: str) -> Tag:
        """Find a Model Version tag, usable only on Model version objects, from its name

        Examples:
            ```python
            tag = client.find_model_version_tag("initial")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.MODEL_VERSION, name)

    @exception_handler
    @beartype
    def find_deployment_tag(self, name: str) -> Tag:
        """Find a Deployment tag, usable only on Deployment objects, from its name

        Examples:
            ```python
            tag = client.find_deployment_tag("operation")
            ```

        Arguments:
            name (str): Name of the tag

        Returns:
            A (Tag) object
        """
        return self._find_tag_organization_scoped(TagTarget.DEPLOYMENT, name)

    @exception_handler
    @beartype
    def get_job_by_id(self, id: UUID | str) -> Job:
        """Get a (Job) from its id.

        Examples:
            ```python
            job = client.get_job_by_id(
                id="YOUR JOB ID"
            )
            ```

        Arguments:
            id (str): job id displayed in your job page.

        Returns:
            A (Job) object
        """
        if isinstance(id, str):
            id = UUID(id)
        params = {"id": id}
        r = self.connexion.get(
            f"/api/organization/{self.id}/jobv2s/find", params=params
        ).json()
        return Job(self.connexion, r, version=2)

    @exception_handler
    @beartype
    def create_processing(
        self,
        name: str,
        type: str | ProcessingType,
        default_cpu: int,
        default_gpu: int,
        default_parameters: dict,
        docker_image: str,
        docker_tag: str,
        docker_flags: list[str] | None = None,
        description: str | None = None,
    ) -> Processing:
        """Create a (Processing) in this organization.

        Examples:
            Create a dataset named datatest with data from datalake and version it
            ```python
            foo_dataset = client.create_processing(
                name="processing-1",
                type="AUTO_TAGGING",
                default_cpu=4,
                default_gpu=0,
                default_parameters={"gamma": 1.2},
                docker_image="training-tf2",
                docker_tag="latest",
                description="A processing to auto-tag images",
            )
            ```

        Arguments:
            name (str): Name of the processing.
            type (str or ProcessingType): Type of the processing.
            default_cpu (int): Default cpu used by this processing.
            default_gpu (int): Default gpu used by this processing.
            default_parameters (dict): Default parameters used by this processing.
            docker_image (str): Docker image of the processing.
            docker_tag (str): Docker tag of the processing.
            docker_flags (list[str], optional): Docker flags of the processing.
            description (str, optional): Description of the processing. Default to None.

        Returns:
            A (Processing)
        """
        type_ = ProcessingType.validate(type)
        payload = {
            "name": name,
            "description": description,
            "type": type_,
            "default_cpu": default_cpu,
            "default_gpu": default_gpu,
            "default_parameters": default_parameters,
            "docker_image": docker_image,
            "docker_tag": docker_tag,
            "docker_flags": docker_flags,
        }
        r = self.connexion.post(
            f"/api/organization/{self.id}/processings", data=orjson.dumps(payload)
        ).json()
        return Processing(self.connexion, r)

    @exception_handler
    @beartype
    def list_processings(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Processing]:
        """List all processings stored in this organization

        Examples:
            ```python
            processings = client.list_processings()
            ```

        Arguments:
            limit (int, optional): Limit number of processings to retrieve. Defaults to None, all models will be retrieved.
            offset (int, optional): Offset to begin with when listing processings. Defaults to None, starting at 0.
            order_by (list[str], optional): Some fields to order processings against. Defaults to None, processings will not be sorted

        Returns:
            A list of all (Processing) that belong to this organization
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = filter_payload(params)
        r = self.connexion.get(
            f"/api/organization/{self.id}/processings", params=params
        ).json()
        return [Processing(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def get_processing(self, name: str) -> Processing:
        """Find a processing in this organization

        Examples:
            ```python
            processing = client.get_processing(name="auto-tagging-dataset")
            ```

        Arguments:
            name (str): Name of the processing you are looking for

        Returns:
            A (Processing) that belong to this organization
        """
        params = {"name": name}
        r = self.connexion.get(
            f"/api/organization/{self.id}/processings/find", params=params
        ).json()
        return Processing(self.connexion, r)

    @exception_handler
    @beartype
    def list_annotation_campaigns(
        self, is_closed: bool | None = None
    ) -> list[AnnotationCampaign]:
        """List annotation campaigns of this organization, created on dataset versions.

        Examples:
            ```python
            campaigns = client.list_annotation_campaigns()
            ```
        Returns:
            A list of (AnnotationCampaign) of this organization
        """
        params = {}
        if is_closed is not None:
            params["is_closed"] = is_closed

        r = self.connexion.get(
            f"/api/organization/{self.id}/annotationcampaigns", params=params
        ).json()
        return [AnnotationCampaign(self.connexion, item) for item in r["items"]]

    @exception_handler
    @beartype
    def list_review_campaigns(
        self, is_closed: bool | None = None
    ) -> list[ReviewCampaign]:
        """List review campaigns of this organization, created on deployments.

        Examples:
            ```python
            campaigns = client.list_review_campaigns()
            ```
        Returns:
            A list of (ReviewCampaign) of this organization
        """
        params = {}
        if is_closed is not None:
            params["is_closed"] = is_closed

        r = self.connexion.get(
            f"/api/organization/{self.id}/reviewcampaigns", params=params
        ).json()
        return [ReviewCampaign(self.connexion, item) for item in r["items"]]



import logging
import os
from pathlib import Path
from uuid import UUID

from orjson import orjson
from PIL import Image

from picsellia.exceptions import UploadError
from picsellia.exif import read_image_metadata
from picsellia.sdk.connexion import Connection
from picsellia.sdk.datasource import DataSource
from picsellia.services.error_manager import ErrorManager
from picsellia.types.enums import DataType, ObjectDataType
from picsellia.utils import filter_payload

ALLOWED_IMAGE_CONTENT_TYPES = (
    "image/jpeg",
    "image/bmp",
    "image/gif",
    "image/png",
    "image/tiff",
)

logger = logging.getLogger("picsellia")


class DataUploader:
    def __init__(
        self,
        connexion: Connection,
        datalake_id: UUID,
        connector_id: UUID | None,
        computed_tag_ids: list[UUID],
        source: DataSource | None,
        fill_metadata: bool,
        upload_dir: str | None,
        error_manager: ErrorManager | None,
    ):
        self.connexion = connexion
        self.connector_id = connector_id
        self.datalake_id = datalake_id
        self.error_manager = error_manager
        self.fill_metadata = fill_metadata
        self.computed_tags_ids = computed_tag_ids
        self.source = source
        self.upload_dir = upload_dir

    def upload(self, items: tuple[str | Path, dict | None, dict | None]):
        path, metadatum, custom_metadatum = items
        try:
            filename = os.path.basename(path)
            object_name = self.connexion.generate_data_object_name(
                filename, ObjectDataType.DATA, self.connector_id, self.upload_dir
            )
            _, _, content_type = self.connexion.upload_file(
                object_name, path, connector_id=self.connector_id
            )
            payload = self._prepare_payload(
                filename, object_name, path, content_type, metadatum, custom_metadatum
            )
            return self.connexion.post(
                f"/api/datalake/{self.datalake_id}/datas",
                data=orjson.dumps(payload),
            )
        except Exception as e:
            message = f"Could not upload path '{path}'"
            logger.error(message, exc_info=e)
            if self.error_manager:
                self.error_manager.append(
                    UploadError(message=message, path=path, parent=e)
                )
            return None

    def _prepare_payload(
        self,
        filename: str,
        object_name: str,
        path: str,
        content_type: str,
        metadatum: dict,
        custom_metadatum: dict | None,
    ):
        meta = None
        metadata = filter_payload(metadatum) if metadatum else None

        if content_type in ALLOWED_IMAGE_CONTENT_TYPES:
            data_type = DataType.IMAGE
            exif_data = self.prepare_metadata_from_image(
                path, metadatum, self.fill_metadata
            )
            if exif_data:
                meta = {"width": exif_data.width, "height": exif_data.height}
                metadata = exif_data.metadata
        elif content_type.startswith("video/"):
            data_type = DataType.VIDEO
        else:
            data_type = DataType.IMAGE
            logger.info(
                f"SDK cannot open {path} because its content type {content_type} is not openable here. "
                f"We will still upload your data and try handle it in our backend services."
            )

        payload = {
            "type": data_type,
            "filename": filename,
            "object_name": object_name,
            "content_type": content_type,
            "tags": self.computed_tags_ids,
            "data_source_id": self.source.id if self.source else None,
        }
        if meta:
            payload["meta"] = meta
        if metadata:
            payload["metadata"] = metadata
        if custom_metadatum:
            payload["custom_metadata"] = custom_metadatum
        return payload

    @staticmethod
    def prepare_metadata_from_image(path: str, metadatum: dict, fill_metadata: bool):
        try:
            # Try opening with Pillow to read exif metadata, width and height
            with Image.open(path) as image:
                return read_image_metadata(image, metadatum, fill_metadata)
        except Exception:
            logger.error(
                f"{path} could not be opened with Pillow, Picsellia will try to compute size of this image later."
            )
            return None



from uuid import UUID

from pydantic import BaseModel, Field

from picsellia.types.enums import (
    AnnotationStatus,
    CampaignStepType,
    DataType,
    DataUploadStatus,
    ExperimentStatus,
    Framework,
    InferenceType,
    JobRunStatus,
    JobStatus,
    LogType,
    ProcessingType,
    TagTarget,
)


class DaoSchema(BaseModel):
    # Might not be defined if refresh is done manually
    id: UUID | None = None


class DatalakeSchema(DaoSchema):
    name: str
    connector_id: UUID | None = None


class OrganizationSchema(DaoSchema):
    name: str
    default_connector_id: UUID | None = None


class DatasetSchema(DaoSchema):
    name: str


class DatasetVersionSchema(DaoSchema):
    name: str = Field(alias="origin_name")
    origin_id: UUID = Field(alias="origin_id")
    annotation_campaign_id: UUID | None = None
    version: str
    type: InferenceType


class AnnotationCampaignSchema(DaoSchema):
    dataset_version_id: UUID


class ReviewCampaignSchema(DaoSchema):
    deployment_id: UUID


class CampaignStepSchema(DaoSchema):
    name: str
    order: int
    type: CampaignStepType


class DatasetVersionStats(BaseModel):
    label_repartition: dict[str, int]
    nb_objects: int
    nb_annotations: int


class ModelSchema(DaoSchema):
    name: str
    private: bool


class ModelVersionSchema(DaoSchema):
    origin: ModelSchema
    name: str
    version: int
    labels: dict | None = None
    type: InferenceType
    framework: Framework


class ModelDataSchema(DaoSchema):
    name: str
    version_id: UUID
    repartition: dict


class ModelContextSchema(DaoSchema):
    experiment_id: UUID | None = None
    datas: list[ModelDataSchema]
    parameters: dict


class ModelFileSchema(DaoSchema):
    name: str
    object_name: str
    filename: str
    large: bool
    url: str | None = Field(alias="presigned_url", default=None)


class ProjectSchema(DaoSchema):
    name: str


class DeploymentSchema(DaoSchema):
    name: str
    type: InferenceType
    oracle_host: str | None = None
    serving_host: str | None = None
    review_campaign_id: UUID | None = None


class TargetDatalakeConnectorSchema(DaoSchema):
    id: UUID
    client_type: str
    bucket_name: str


class ImageMetaSchema(BaseModel):
    width: int
    height: int


class VideoMetaSchema(BaseModel):
    width: int
    height: int


class DataSchema(DaoSchema):
    object_name: str
    filename: str
    type: DataType
    content_type: str
    upload_status: DataUploadStatus
    url: str | None = Field(alias="presigned_url", default=None)
    metadata: dict | None = None
    custom_metadata: dict | None = None


class ImageSchema(DataSchema):
    meta: ImageMetaSchema


class VideoSchema(DataSchema):
    meta: VideoMetaSchema


class DataProjectionSchema(DaoSchema):
    name: str | None = None
    object_name: str | None = None
    filename: str | None = None
    compute_status: DataUploadStatus
    url: str | None = Field(alias="presigned_url", default=None)
    infos: dict | None = None


class DataSourceSchema(DaoSchema):
    name: str


class AssetSchema(DaoSchema):
    data: ImageSchema | VideoSchema


class PredictedAssetSchema(DaoSchema):
    data: ImageSchema
    oracle_prediction_id: UUID


class ExperimentSchema(DaoSchema):
    name: str
    project_id: UUID
    status: ExperimentStatus


class FastTrainingSchema(DaoSchema):
    job_id: UUID
    job_run_id: UUID
    status: JobRunStatus
    experiment_id: UUID
    experiment_name: str


class EvaluationSchema(DaoSchema):
    asset_id: UUID


class UserSchema(DaoSchema):
    username: str


class WorkerSchema(DaoSchema):
    username: str
    user_id: UUID


class ArtifactSchema(DaoSchema):
    name: str
    object_name: str
    filename: str
    large: bool
    url: str | None = Field(alias="presigned_url", default=None)


class LoggingFileSchema(DaoSchema):
    object_name: str
    url: str | None = Field(alias="presigned_url", default=None)


LogDataType = list | dict | float | int | str


class LogSchema(DaoSchema):
    name: str
    type: LogType
    data: LogDataType


class TagSchema(DaoSchema):
    name: str
    target_type: TagTarget


class LabelSchema(DaoSchema):
    name: str
    group_id: UUID | None = None


class LabelGroupSchema(DaoSchema):
    name: str
    parent_id: UUID | None = None


class AnnotationSchema(DaoSchema):
    duration: float
    status: AnnotationStatus


class ShapeSchema(DaoSchema):
    label: LabelSchema
    text: str | None = None


class RectangleSchema(ShapeSchema):
    x: int
    y: int
    w: int
    h: int


class PolygonSchema(ShapeSchema):
    coords: list = Field(alias="polygon")


class LineSchema(ShapeSchema):
    coords: list = Field(alias="line")


class PointSchema(ShapeSchema):
    coords: list = Field(alias="point")


class ClassificationSchema(ShapeSchema):
    pass


class JobSchema(DaoSchema):
    status: JobStatus


class JobRunSchema(DaoSchema):
    status: JobRunStatus


class ProcessingSchema(DaoSchema):
    name: str
    type: ProcessingType
    docker_image: str
    docker_tag: str


# helpers
class CloudProjectionObject(BaseModel):
    name: str
    object_name: str


class CloudObject(BaseModel):
    metadata: dict | None = None
    custom_metadata: dict | None = None
    tags: list[str] = Field(default_factory=list)
    data_source: str | None = None



import logging
from pathlib import Path

from beartype import beartype
from picsellia_annotations.coco import Annotation as COCOAnnotation
from picsellia_annotations.coco import COCOFile
from picsellia_annotations.exceptions import FileError, ParsingError
from picsellia_annotations.utils import read_coco_file, read_video_coco_file
from picsellia_annotations.video_coco import (
    ExtendedImage,
    VideoAnnotation,
    VideoCOCOFile,
)

from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    FileNotFoundException,
    UnparsableAnnotationFileException,
)
from picsellia.sdk.label import Label
from picsellia.types.enums import InferenceType

logger = logging.getLogger("picsellia")


@exception_handler
@beartype
def parse_coco_file(file_path: Path | str) -> COCOFile:
    try:
        return read_coco_file(file_path=file_path)
    except FileError:
        raise FileNotFoundException(f"{file_path} not found")
    except ParsingError as e:
        raise UnparsableAnnotationFileException(
            f"Could not parse COCO file {file_path} because : {e}"
        )


@exception_handler
@beartype
def parse_coco_video_file(file_path: Path | str) -> VideoCOCOFile:
    try:
        return read_video_coco_file(file_path=file_path)
    except FileError:
        raise FileNotFoundException(f"{file_path} not found")
    except ParsingError as e:
        raise UnparsableAnnotationFileException(
            f"Could not parse COCO VID file {file_path} because : {e}"
        )


@exception_handler
@beartype
def read_annotations(  # noqa: C901
    cocofile_type: InferenceType,
    coco_annotations: list[COCOAnnotation],
    labels: dict[int, Label],
    assets: dict[int, str],
) -> list[dict]:
    annotations_asset_map: dict[str, dict] = {}
    for annotation in coco_annotations:
        try:
            label = labels[annotation.category_id]
        except KeyError:  # pragma: no cover
            logger.error(
                f"category_id {annotation.category_id} not found into retrieved labels"
            )
            continue

        try:
            asset_id = assets[annotation.image_id]
        except KeyError:
            logger.error(
                f"image_id {annotation.image_id} not found into retrieved assets"
            )
            continue

        if asset_id not in annotations_asset_map:
            annotations_asset_map[asset_id] = {
                "rectangles": [],
                "classifications": [],
                "polygons": [],
                "keypoints": [],
            }

        if cocofile_type == InferenceType.SEGMENTATION:
            if not annotation.is_rle():
                polygon_coords = annotation.polygon_to_list_coordinates()
            else:
                logger.error(
                    f"annotation_id {annotation.id} is a RLE which is not supported yet"
                )
                continue

            for polygon_coord in polygon_coords:
                annotations_asset_map[asset_id]["polygons"].append(
                    {
                        "polygon": polygon_coord,
                        "label_id": label.id,
                        "text": annotation.utf8_string,
                    }
                )
        elif cocofile_type == InferenceType.OBJECT_DETECTION:
            annotations_asset_map[asset_id]["rectangles"].append(
                {
                    "x": int(annotation.bbox[0]),
                    "y": int(annotation.bbox[1]),
                    "w": int(annotation.bbox[2]),
                    "h": int(annotation.bbox[3]),
                    "label_id": label.id,
                    "text": annotation.utf8_string,
                }
            )
        elif cocofile_type == InferenceType.CLASSIFICATION:
            annotations_asset_map[asset_id]["classifications"].append(
                {"label_id": label.id, "text": annotation.utf8_string}
            )
        elif cocofile_type == InferenceType.KEYPOINT:
            if len(annotation.keypoints) % 3 != 0:
                logging.error(
                    f"Could not parse keypoints of annotation_id {annotation.id}, it must be a list like [x1, y1, v1, x2, y2, v2...]"
                )
                continue

            keypoints = []
            for i in range(0, len(annotation.keypoints), 3):
                batch = annotation.keypoints[i : i + 3]
                keypoints.append([batch[0], batch[1], int(batch[2])])
            annotations_asset_map[asset_id]["keypoints"].append(
                {
                    "label_id": label.id,
                    "text": annotation.utf8_string,
                    "keypoints": keypoints,
                }
            )

    return [
        {"asset_id": asset_id, **annotation}
        for asset_id, annotation in annotations_asset_map.items()
    ]


@exception_handler
@beartype
def read_video_annotations(  # noqa: C901
    cocofile_type: InferenceType,
    coco_annotations: list[VideoAnnotation],
    coco_frames: list[ExtendedImage],
    labels: dict[int, Label],
    assets: dict[int, str],
    only_key_frames: bool = True,
    use_image_id_as_frame_id: bool = False,
) -> list[dict]:
    frames: dict[int, ExtendedImage] = {
        coco_frame.id: coco_frame for coco_frame in coco_frames
    }

    tracks: dict[tuple[int, int], dict] = {}
    for annotation in coco_annotations:
        if annotation.attributes is None:
            logger.warning(
                f"annotation {annotation.id} does not have attributes with track_id information. We cannot parse it."
            )
            continue

        if only_key_frames and annotation.attributes.keyframe is False:
            # If you consider only key frames of COCO file, ignore others
            logger.debug(
                f"skipping annotation {annotation.id} because it is not a keyframe"
            )
            continue

        if annotation.attributes.track_id is None:
            logger.warning(
                f"annotation {annotation.id} has no 'attributes.track_id' set, we cannot infer track"
            )
            continue

        coco_track_id = annotation.attributes.track_id
        track_key = (annotation.video_id, coco_track_id)

        if annotation.attributes.frame_id is None:
            if annotation.image_id not in frames:
                logger.warning(
                    f"annotation {annotation.id} has no 'attributes.frame_id' and its image_id is unknown, we cannot infer frame index"
                )
                continue
            elif frames[annotation.image_id].frame_id is None:
                if use_image_id_as_frame_id:
                    frame_index = annotation.image_id
                else:
                    logger.warning(
                        f"annotation {annotation.id} is bound to image {annotation.image_id} which does not have frame_id, so we cannot infer frame index"
                    )
                    continue
            else:
                frame_index = frames[annotation.image_id].frame_id
        else:
            frame_index = annotation.attributes.frame_id

        try:
            label = labels[annotation.category_id]
        except KeyError:  # pragma: no cover
            logger.error(
                f"category_id {annotation.category_id} not found into retrieved labels"
            )
            continue

        if track_key not in tracks:
            tracks[track_key] = {
                "label_id": label.id,
                "start_frame_index": frame_index,
                # This field is not used by api
                "_last_frame_index": frame_index + 1,
                "polygon_keyframes": [],
                "rectangle_keyframes": [],
            }
        elif frame_index < tracks[track_key]["start_frame_index"]:
            # Case where current annotation is before current first keyframe of track
            # So we will store new start frame index
            # and we also need to reset already computed relative frame indexes
            offset = tracks[track_key]["start_frame_index"] - frame_index
            tracks[track_key]["start_frame_index"] = frame_index
            for keyframe_type in ["rectangle_keyframes", "polygon_keyframes"]:
                for keyframe in tracks[track_key][keyframe_type]:
                    keyframe["frame_index"] = keyframe["frame_index"] + offset

        # Only take first utf8_string key we find in annotations
        if "text" not in tracks[track_key] and annotation.utf8_string is not None:
            tracks[track_key]["text"] = annotation.utf8_string

        # Picsellia accepts relative frame index as frame_index of keyframes
        start_frame_index = tracks[track_key]["start_frame_index"]
        relative_frame_index = frame_index - start_frame_index

        if cocofile_type == InferenceType.SEGMENTATION:
            if not annotation.is_rle():
                polygon_coords = annotation.polygon_to_list_coordinates()
            else:
                logger.error(
                    f"annotation_id {annotation.id} is a RLE which is not supported yet"
                )
                continue

            for polygon_coord in polygon_coords:
                tracks[track_key]["polygon_keyframes"].append(
                    {"polygon": polygon_coord, "frame_index": relative_frame_index}
                )
        elif cocofile_type == InferenceType.OBJECT_DETECTION:
            tracks[track_key]["rectangle_keyframes"].append(
                {
                    "x": int(annotation.bbox[0]),
                    "y": int(annotation.bbox[1]),
                    "w": int(annotation.bbox[2]),
                    "h": int(annotation.bbox[3]),
                    "frame_index": relative_frame_index,
                }
            )
        elif cocofile_type == InferenceType.CLASSIFICATION:
            raise NotImplementedError()

        tracks[track_key]["_last_frame_index"] = max(
            tracks[track_key]["_last_frame_index"], frame_index + 1
        )

    annotations_asset_map: dict[str, dict] = {}
    for track_key, track in tracks.items():
        video_id, _ = track_key
        try:
            asset_id = assets[video_id]
        except KeyError:
            logger.error(f"image_id {video_id} not found into retrieved assets")
            continue

        if asset_id not in annotations_asset_map:
            annotations_asset_map[asset_id] = {"tracks": []}

        # End key frames are relative
        relative_frame_index = track["_last_frame_index"] - track["start_frame_index"]
        track["end_keyframes"] = [{"frame_index": relative_frame_index}]

        annotations_asset_map[asset_id]["tracks"].append(track)

    return [
        {"asset_id": asset_id, **annotation}
        for asset_id, annotation in annotations_asset_map.items()
    ]



class PicselliaError(Exception):
    """Base class for exceptions."""

    def __init__(self, message="Something went wrong"):
        """
        Arguments:
            message (str): Informative message about the exception.
        """
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class AuthenticationError(PicselliaError):
    """Raised by picsellia-tf2 package"""

    pass


class BadConfigurationContinuousTrainingError(PicselliaError):
    """Raised when a setup of continuous training is not well configured"""

    pass


class BadRequestError(PicselliaError):
    """Indicates a malformed or unsupported query. This can be the result of either client
    or server side query validation."""

    pass


class RequestTooLargeError(PicselliaError):
    """Request is too large and server do not allow it. You might want to retry with a smaller query or payload"""

    pass


class BadGatewayError(PicselliaError):
    """Picsellia is unavailable at the moment"""

    pass


class TooManyRequestError(PicselliaError):
    """Raised when too much request are done to the server."""

    pass


class ContextSourceNotDefined(PicselliaError):
    """Raised if experiment of a ModelContext is not defined"""

    pass


class ContextDataNotDefined(PicselliaError):
    """Raised if dataset of a ModelContext is not defined"""

    pass


class DistantStorageError(PicselliaError):
    """Raised when an upload or a download from S3 storage went wrong"""

    pass


class UploadError(PicselliaError):
    """Raised when an upload went wrong"""

    def __init__(self, message: str, path: str, parent: Exception | None):
        super().__init__(message)
        self.path = path
        self.parent = parent


class UnprocessableData(UploadError):
    """Raised when a data could not be processed by our services"""

    def __init__(self, data):
        super().__init__(
            message=f"{data.filename} could not be processed by our services.",
            path=data.filename,
            parent=None,
        )


class DownloadError(PicselliaError):
    """Raised when a download went wrong"""

    pass


class FileNotFoundException(PicselliaError):
    """Raised when a file is not found"""

    pass


class ForbiddenError(PicselliaError):
    """Raised when your token does not match to any known token"""

    pass


class InsufficientResourcesError(PicselliaError):
    """Raised when your token does not match to any known token"""

    pass


class InternalServerError(PicselliaError):
    """Raised when Picsellia threw an unusual InternalServerError"""

    pass


class MonitorError(PicselliaError):
    """Raised when a prediction could not have been monitored by Oracle"""

    pass


class NetworkError(PicselliaError):
    """Raised when an HTTPError occurs."""

    pass


class NoBaseExperimentError(PicselliaError):
    """Raised when exception has no base experiment"""

    pass


class NoBaseModelVersionError(PicselliaError):
    """Raised when exception has no base model"""

    pass


class NoConnectorFound(PicselliaError):
    """Raised if connexion object is not pointed to any organization"""

    pass


class NoDataError(PicselliaError):
    """Raised when you try to retrieve data from an empty datalake"""

    pass


class NoShadowModel(PicselliaError):
    """Raised when there is no shadow model for a deployment"""

    pass


class NothingDoneError(PicselliaError):
    """Raised when something should have been done but nothing happened"""

    pass


class PredictionError(PicselliaError):
    """Raised when a prediction could not have been done"""

    pass


class ResourceConflictError(PicselliaError):
    """Exception raised when a given resource already exists."""

    pass


class ResourceLockedError(PicselliaError):
    """Exception raised when a resource is locked, and you're trying to create, update or delete a linked resource"""

    pass


class ResourceNotFoundError(PicselliaError):
    """Exception raised when a given resource is not found."""

    pass


class UnauthorizedError(PicselliaError):
    """Raised when your token does not match to any known token"""

    pass


class UndefinedObjectError(PicselliaError):
    """Raised when Dao is initialized without id"""

    pass


class AnnotationFileIncoherentTypeException(PicselliaError):
    """Raised when annotation file can't be imported in a dataset due to its type"""

    pass


class UnparsableAnnotationFileException(PicselliaError):
    """Raised when annotation file is unparsable"""

    pass


class UploadFailed(PicselliaError):
    """Raised when an upload of a file has failed"""

    pass


class WaitingAttemptsTimeout(PicselliaError):
    """Raised when a job.wait_for_status is taking too much attempts"""

    pass


class WrongJobVersionError(PicselliaError):
    """Raised when a user attenmpt to use a JobV2 method on a legacy job"""

    pass


class IllegalJobTransitionError(PicselliaError):
    """Raised when you try to change a job status to an incoherent one"""

    pass


class NotSupportedJobVersionError(PicselliaError):
    """Raised when you try to use an unsupported job version"""

    pass


class NoUrlAvailable(PicselliaError):
    """Raised on download when no url could be retrieved from platform"""

    pass


class ContentTypeUnknown(PicselliaError):
    """Raised when content type is not known on monitor"""

    pass


class MonitoringConnectionError(PicselliaError):
    """Raised when a call to Oracle or Serving is returning a 50X"""

    pass



import logging
from uuid import UUID

from beartype import beartype

import picsellia.exceptions as exceptions
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.dataset_version import DatasetVersion
from picsellia.types.schemas import ModelContextSchema, ModelDataSchema

logger = logging.getLogger("picsellia")


class ModelContext(Dao):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def experiment_id(self) -> UUID:
        """UUID of the original (Experiment) that generated this (Model)"""
        return self._experiment_id

    @property
    def datasets(self) -> list[ModelDataSchema] | None:
        """List of (ModelDataSchema) with ids, names, and repartition of (DatasetVersion) used by original (Experiment)"""
        return self._datasets

    @property
    def parameters(self) -> dict:
        """Parameters used by original (Experiment) to generated (Model)"""
        return self._parameters

    def __str__(self):
        return f"A {Colors.BLUE}model context{Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/model/context/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> ModelContextSchema:
        schema = ModelContextSchema(**data)
        self._experiment_id = schema.experiment_id
        self._datasets = schema.datas
        self._parameters = schema.parameters
        return schema

    @exception_handler
    @beartype
    def get_infos(self) -> dict:
        """Retrieve some infos about this context

        Examples:
            ```python
            infos = model_context.get_infos()
            ```

        Returns:
            A dict with experiment, datasets and parameters
        """
        return {
            "experiment": self.experiment_id,
            "datasets": self.datasets,
            "parameters": self.parameters,
        }

    @exception_handler
    @beartype
    def get_experiment(self):
        """Retrieve source experiment of this context

        It will raise an exception if this context has no experiment source

        Examples:
            ```python
            exp = model_context.get_experiment()
            ```

        Returns:
            An (Experiment) object
        """
        if self.experiment_id is None:
            raise exceptions.ContextSourceNotDefined(
                "This context has no experiment source"
            )

        from .experiment import Experiment

        r = self.connexion.get(f"/api/experiment/{self.experiment_id}").json()
        return Experiment(self.connexion, r)

    @exception_handler
    @beartype
    def get_dataset_version(self, name: str) -> DatasetVersion:
        """Retrieve dataset version used to train or evaluate the model, by the name given when attached experiment and dataset.

        It will raise an exception if this context has no dataset attached with given name.

        Examples:
            ```python
            dataset_version = model_context.get_dataset_version("train")
            ```

        Arguments:
            name (str): Name of the dataset version attached to the experiment

        Returns:
            A (DatasetVersion) object
        """
        if self._datasets is None or self._datasets == []:
            raise exceptions.ContextDataNotDefined("This context has no data")

        for data in self._datasets:
            if data.name == name:
                r = self.connexion.get(f"/api/dataset/version/{data.version_id}").json()
                return DatasetVersion(self.connexion, r)

        raise exceptions.ContextDataNotDefined(
            f"This context has no dataset attached with {name}"
        )

    @exception_handler
    @beartype
    def retrieve_datasets(
        self,
    ) -> dict[str, DatasetVersion]:
        """Retrieve datasets used to train and evaluate (or else) the model

        It will raise an exception if this context has no data

        Examples:
            ```python
            dataset_versions = model_context.retrieve_datasets()
            ```

        Returns:
            A dict with dataset name as key and (DatasetVersion) as value
        """
        if self._datasets is None or self._datasets == []:
            raise exceptions.ContextSourceNotDefined("This context has no data")

        datasets = {}
        for dataset_version in self._datasets:
            r = self.connexion.get(
                f"/api/dataset/version/{dataset_version.version_id}"
            ).json()
            datasets[dataset_version.name] = DatasetVersion(self.connexion, r)
        return datasets



from urllib.parse import urljoin
from uuid import UUID

import requests
from orjson import orjson
from requests import Session

from picsellia.decorators import retry


class PlatformAdapter:
    def __init__(self, connector_id: UUID, session: Session, host: str, headers: dict):
        self.connector_id = connector_id
        self.host = host
        self.session = session
        self.headers = headers

    @retry(requests.ConnectionError)
    def generate_presigned_url(
        self, object_name: str, content_type: str
    ) -> requests.Response:
        payload = {"object_name": object_name, "content_type": content_type}
        path = f"/api/object-storage/{self.connector_id}/generate_presigned_url"
        url = urljoin(self.host, path)
        response = self.session.post(
            url=url, data=orjson.dumps(payload), headers=self.headers
        )
        response.raise_for_status()
        return response

    @retry(requests.ConnectionError)
    def init_multipart_upload(
        self, object_name: str, content_type: str
    ) -> requests.Response:
        payload = {"object_name": object_name, "content_type": content_type}
        path = f"/api/object-storage/{self.connector_id}/init_multipart_upload"
        url = urljoin(self.host, path)
        response = self.session.post(
            url=url, data=orjson.dumps(payload), headers=self.headers
        )
        response.raise_for_status()
        return response

    @retry(requests.ConnectionError)
    def generate_part_presigned_url(
        self, object_name: str, additional_payload: dict
    ) -> requests.Response:
        payload = {"object_name": object_name, **additional_payload}
        path = f"/api/object-storage/{self.connector_id}/generate_part_presigned_url"
        url = urljoin(self.host, path)
        response = self.session.post(
            url=url, data=orjson.dumps(payload), headers=self.headers
        )
        response.raise_for_status()
        return response

    @retry(requests.ConnectionError)
    def complete_part_upload(
        self, object_name: str, additional_payload: dict
    ) -> requests.Response:
        payload = {"object_name": object_name, **additional_payload}
        path = f"/api/object-storage/{self.connector_id}/complete_part_upload"
        url = urljoin(self.host, path)
        response = self.session.post(
            url=url, data=orjson.dumps(payload), headers=self.headers
        )
        response.raise_for_status()
        return response



import logging
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.types.schemas import LabelGroupSchema

logger = logging.getLogger("picsellia")


class LabelGroup(Dao):
    def __init__(
        self, connexion: Connection, dataset_version_id: UUID, data: dict
    ) -> None:
        Dao.__init__(self, connexion, data)
        self._dataset_version_id = dataset_version_id

    @property
    def name(self) -> str:
        """Name of this (LabelGroup)"""
        return self._name

    @property
    def parent_id(self) -> UUID | None:
        """Id of the (LabelGroup) parent of this (LabelGroup)"""
        return self._parent_id

    @property
    def dataset_version_id(self) -> UUID:
        """Id of the (DatasetVersion) of this (LabelGroup)"""
        return self._dataset_version_id

    def __str__(self):
        return f"{Colors.YELLOW}LabelGroup '{self.name}'{Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/label/group/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> LabelGroupSchema:
        schema = LabelGroupSchema(**data)
        self._name = schema.name
        self._parent_id = schema.parent_id
        return schema

    @exception_handler
    @beartype
    def update(self, name: str = None, parent_id: UUID = None) -> None:
        """Update this (LabelGroup) with a new name or a new parent_id.

        Examples:
            ```python
            group.update(name="new")
            ```

        Arguments:
            name: New name of this (LabelGroup)
            parent_id (uuid, optional): New name of this (LabelGroup)
        """
        payload = {}
        if name:
            payload["name"] = name

        if parent_id:
            payload["parent_id"] = parent_id

        r = self.connexion.patch(
            f"/api/label/group/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def set_parent(self, parent: "LabelGroup") -> None:
        """Set a new parent for this (LabelGroup)

        Examples:
            ```python
            root = dataset_version.create_label_group("root")
            leaf = dataset_version.create_label_group("leaf")
            leaf.set_parent(root)
            ```

        Arguments:
            parent (LabelGroup): New parent of this (LabelGroup)
        """
        if parent.dataset_version_id != self.dataset_version_id:
            raise ValueError("LabelGroup must belong to the same DatasetVersion")
        return self.update(parent_id=parent.id)

    @exception_handler
    @beartype
    def get_parent(self) -> "LabelGroup | None":
        """Get (LabelGroup) parent of this (LabelGroup). If no (LabelGroup) is set, return None.

        Examples:
            ```python
            root = dataset_version.create_label_group("root")
            leaf = dataset_version.create_label_group("leaf", root)
            group = leaf.get_parent()
            ```

        Returns:
            a (LabelGroup) if there is a parent, else None
        """
        if not self._parent_id:
            return None

        r = self.connexion.get(f"/api/label/group/{self._parent_id}").json()
        return LabelGroup(self.connexion, self._dataset_version_id, r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this (LabelGroup) from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            group.delete()
            ```
        """
        self.connexion.delete(f"/api/label/group/{self.id}")
        logger.info(f"{self} deleted.")



import logging
from uuid import UUID

from beartype import beartype

from picsellia import exceptions
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.multi_object import MultiObject
from picsellia.types.schemas import EvaluationSchema

logger = logging.getLogger("picsellia")


class Evaluation(Dao):
    def __init__(self, connexion: Connection, context_id: UUID, data: dict) -> None:
        Dao.__init__(self, connexion, data)
        self._context_id = context_id

    @property
    def asset_id(self) -> UUID:
        """UUID of the (Asset) of this (Evaluation)"""
        return self._asset_id

    @property
    def context_id(self) -> UUID:
        """UUID of the context (it can be an experiment or a model version) of this (Evaluation)"""
        return self._context_id

    def __str__(self):
        return f"{Colors.BLUE}Evaluation of {self.asset_id} from context {self.context_id} {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/evaluation/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> EvaluationSchema:
        schema = EvaluationSchema(**data)
        self._asset_id = schema.asset_id
        return schema

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this evaluation from the platform.
        All evaluated shapes will be deleted!
        This is a very dangerous move.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            one_evaluation.delete()
            ```
        """
        self.connexion.delete(f"/api/evaluation/{self.id}")
        logger.info(f"{self} deleted from platform.")


class MultiEvaluation(MultiObject[Evaluation]):
    @beartype
    def __init__(
        self, connexion: Connection, context_id: UUID, items: list[Evaluation]
    ):
        MultiObject.__init__(self, connexion, items)
        self._context_id = context_id

    @property
    def context_id(self) -> UUID:
        return self._context_id

    def __str__(self) -> str:
        return f"{Colors.GREEN}MultiEvaluation for context {self.context_id}{Colors.ENDC}  size: {len(self)}"

    def __getitem__(self, key) -> "Evaluation | MultiEvaluation":
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.items)))
            evaluations = [self.items[i] for i in indices]
            return MultiEvaluation(self.connexion, self.context_id, evaluations)
        return self.items[key]

    @beartype
    def __add__(self, other) -> "MultiEvaluation":
        self.assert_same_connexion(other)
        items = self.items.copy()
        self._add_other_items_to_items(other, items)
        return MultiEvaluation(self.connexion, self.context_id, items)

    @beartype
    def __iadd__(self, other) -> "MultiEvaluation":
        self.assert_same_connexion(other)
        self._add_other_items_to_items(other, self.items)
        return self

    def _add_other_items_to_items(self, other, items) -> None:
        if isinstance(other, MultiEvaluation):
            if other.context_id != self.context_id:
                raise exceptions.BadRequestError(
                    "These evaluations does not come from the same context"
                )
            items.extend(other.items.copy())
        elif isinstance(other, Evaluation):
            if other.context_id != self.context_id:
                raise exceptions.BadRequestError(
                    "This evaluation does not come from the same context"
                )
            items.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

    def copy(self) -> "MultiEvaluation":
        return MultiEvaluation(self.connexion, self.context_id, self.items.copy())



from pydantic import BaseModel, Field, model_validator

from picsellia.types.enums import InferenceType


class PredictionFormat(BaseModel):
    detection_classes: list[int]
    detection_texts: list[str] | None = None

    @property
    def model_type(cls) -> InferenceType:  # pragma: no cover
        raise NotImplementedError()


class ClassificationPredictionFormat(PredictionFormat):
    detection_scores: list[float]

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.CLASSIFICATION

    @model_validator(mode="before")
    @classmethod
    def check_sizes(cls, data):
        labels, scores = (
            data.get("detection_classes"),
            data.get("detection_scores"),
        )

        if labels is None or scores is None or len(labels) != len(scores):
            raise ValueError("incoherent lists")

        texts = data.get("detection_texts")
        if texts and len(texts) != len(labels):
            raise ValueError(
                "texts are not well defined, there must be exactly the same count of texts than classes"
            )

        return data


class DetectionPredictionFormat(PredictionFormat):
    detection_boxes: list[list[int]]
    detection_scores: list[float]

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.OBJECT_DETECTION

    @model_validator(mode="before")
    @classmethod
    def check_sizes(cls, data):
        labels, scores, boxes = (
            data.get("detection_classes"),
            data.get("detection_scores"),
            data.get("detection_boxes"),
        )

        if (
            labels is None
            or scores is None
            or boxes is None
            or len(labels) != len(scores)
            or len(boxes) != len(labels)
        ):
            raise ValueError("incoherent lists")

        texts = data.get("detection_texts")
        if texts and len(texts) != len(labels):
            raise ValueError(
                "texts are not well defined, there must be exactly the same count of texts and classes"
            )

        return data


class SegmentationPredictionFormat(PredictionFormat):
    detection_scores: list[float]
    detection_masks: list[list[list[int]]]

    detection_boxes: list[list[int]] = Field(
        default=None, exclude=True, deprecated=True
    )

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.SEGMENTATION

    @model_validator(mode="before")
    @classmethod
    def check_sizes(cls, data):
        labels, scores, masks = (
            data.get("detection_classes"),
            data.get("detection_scores"),
            data.get("detection_masks"),
        )

        if (
            labels is None
            or scores is None
            or masks is None
            or len(labels) != len(scores)
            or len(masks) != len(labels)
        ):
            raise ValueError("incoherent lists")

        if any(len(point) != 2 for mask in masks for point in mask):
            raise ValueError("masks' point must contains exactly 2 values")

        texts = data.get("detection_texts")
        if texts and len(texts) != len(labels):
            raise ValueError(
                "texts are not well defined, there must be exactly the same count of texts than classes"
            )

        return data


class LinePredictionFormat(PredictionFormat):
    detection_scores: list[float]
    lines: list[list[list[int]]]

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.LINE

    @model_validator(mode="before")
    @classmethod
    def check_sizes(cls, data):
        labels, scores, lines = (
            data.get("detection_classes"),
            data.get("detection_scores"),
            data.get("lines"),
        )

        if (
            labels is None
            or scores is None
            or lines is None
            or len(labels) != len(scores)
            or len(lines) != len(labels)
        ):
            raise ValueError("incoherent lists")

        if any(len(point) != 2 for line in lines for point in line):
            raise ValueError("lines's point must contains exactly 2 values")

        texts = data.get("detection_texts")
        if texts and len(texts) != len(labels):
            raise ValueError(
                "texts are not well defined, there must be exactly the same count of texts than classes"
            )

        return data


class PointPredictionFormat(PredictionFormat):
    detection_scores: list[float]
    points: list[list[int]]

    @property
    def model_type(cls) -> InferenceType:
        return InferenceType.POINT

    @model_validator(mode="before")
    @classmethod
    def check_sizes(cls, data):
        labels, scores, points = (
            data.get("detection_classes"),
            data.get("detection_scores"),
            data.get("points"),
        )

        if (
            labels is None
            or scores is None
            or points is None
            or len(labels) != len(scores)
            or len(points) != len(labels)
        ):
            raise ValueError("incoherent lists")

        if any(len(point) != 2 for point in points):
            raise ValueError("points must contains exactly 2 values")

        texts = data.get("detection_texts")
        if texts and len(texts) != len(labels):
            raise ValueError(
                "texts are not well defined, there must be exactly the same count of texts than classes"
            )

        return data



import logging
import os

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.downloadable import Downloadable
from picsellia.types.schemas import LoggingFileSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class LoggingFile(Dao, Downloadable):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)
        Downloadable.__init__(self)

    def __str__(self):
        return f"{Colors.GREEN}Logging file {self.object_name}{Colors.ENDC} (id: {self.id})"

    @property
    def object_name(self) -> str:
        """Object name of this (LoggingFile)"""
        return self._object_name

    @property
    def filename(self) -> str:
        """Filename of this (LoggingFile)."""
        return os.path.basename(self._object_name)

    @property
    def large(self) -> bool:
        """(LoggingFile) are usually not large file."""
        return False

    @exception_handler
    @beartype
    def reset_url(self) -> str:
        """Reset url property of this LoggingFile by calling platform.

        Returns:
            A url as a string of this LoggingFile.
        """
        self._url = self.connexion.init_download(self.object_name)
        return self._url

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> LoggingFileSchema:
        schema = LoggingFileSchema(**data)
        self._object_name = schema.object_name
        self._url = schema.url
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/logging/file/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def update(
        self,
        object_name: str | None = None,
    ) -> None:
        """Update this artifact.

        Examples:
            ```python
            this_artifact.update(object_name="another-path-to-artifact")
            ```

        Arguments:
            object_name (str, optional): New object_name of this artifact. Defaults to None.
        """
        payload = {
            "object_name": object_name,
        }
        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"/api/logging/file/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this artifact

        Examples:
            ```python
            this_artifact.delete()
            ```
        """
        self.connexion.delete(f"/api/logging/file/{self.id}")



import logging
import os
from datetime import date
from uuid import UUID

from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.campaign.abstract_campaign import AbstractCampaign
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.job import Job
from picsellia.sdk.review_campaign_step import ReviewCampaignStep
from picsellia.types.enums import CampaignStepType, ObjectDataType, WorkerType
from picsellia.types.schemas import ReviewCampaignSchema

logger = logging.getLogger("picsellia")


class ReviewCampaign(AbstractCampaign[ReviewCampaignStep]):
    _base_path = "/api/campaigns/review"
    _allowed_step_types = [CampaignStepType.REVIEW, CampaignStepType.QUALITY_CONTROL]
    _worker_type = WorkerType.DEPLOYMENT

    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)

    def __str__(self):
        return f"{Colors.GREEN}Review Campaign {Colors.ENDC} for deployment {self.deployment_id} (id: {self.id})"

    @property
    def deployment_id(self) -> UUID:
        """UUID of the (Deployment) of this campaign"""
        return self._deployment_id

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = ReviewCampaignSchema(**data)
        self._deployment_id = schema.deployment_id
        return schema

    @exception_handler
    @beartype
    def update(
        self,
        description: str | None = None,
        instructions_object_name: str | None = None,
        instructions_text: str | None = None,
        end_date: date | None = None,
        auto_add_new_assets: bool | None = None,
        auto_close_on_completion: bool | None = None,
        default_entry_step: ReviewCampaignStep | None = None,
    ) -> None:
        """Update this campaign parameters

        Examples:
            ```python
            foo_campaign.update(description="Yet another campaign")
            ```

        Arguments:
            description (str, optional): Description of the campaign. Defaults to None.
            instructions_object_name (str, optional): Instructions file object name stored on S3. Defaults to None.
            instructions_text (str, optional): Instructions text. Defaults to None.
            end_date (date, optional): End date of the campaign. Defaults to None.
            auto_add_new_assets (bool, optional):
                If true, new assets of this dataset will be added as a task in the campaign.  Defaults to None.
            auto_close_on_completion (bool, optional):
                If true, campaign will be close when all tasks will be done. Defaults to None.
            default_entry_step (Step, optional): Step where tasks will be created for new assets.
        """
        return self._update(
            description,
            instructions_object_name,
            instructions_text,
            end_date,
            auto_add_new_assets,
            auto_close_on_completion,
            default_entry_step,
            annotated_entry_step=None,
        )

    @exception_handler
    @beartype
    def upload_instructions_file(self, path: str) -> None:
        """Upload instructions for this campaign

        Examples:
            ```python
            foo_campaign.upload_instructions_file("/path/to/file.pdf")
            ```

        Arguments:
            path (str): Path of instructions file

        """
        instruction_file_name = os.path.basename(path)
        object_name = self.connexion.generate_deployment_object_name(
            instruction_file_name,
            ObjectDataType.REVIEW_CAMPAIGN_FILE,
            deployment_id=self.deployment_id,
        )
        self.connexion.upload_file(object_name, path)
        self.update(instructions_object_name=object_name)

    def _build_step(self, data: dict):
        return ReviewCampaignStep(self.connexion, data, self._id)

    @exception_handler
    @beartype
    def sync_assets(self) -> Job:
        """This will create tasks and assignments for assets that are not in the campaign yet.

        Your already reviewed assets will NOT be added to this campaign.

        Returns:
            a Job, that you can wait for.
        """
        return self._sync_assets(None)

    @exception_handler
    @beartype
    def reopen(self) -> None:
        """Reopen a closed campaign.

        Examples:
            ```python
            foo_campaign.reopen()
            ```
        """
        self.connexion.post(f"{self._base_path}/{self.id}/reopen")
        logger.info(f"{self} reopened")
